use candle::{DType, Device, Tensor};
use candle_nn::{
    linear,
    loss::mse,
    optim::{AdamW, Optimizer, ParamsAdamW},
    Linear, Module, VarBuilder, VarMap,
};
use std::error::Error;
use tqdm::tqdm;

struct DenseNeuralNetwork {
    ln1: Linear,
    ln2: Linear,
}

impl DenseNeuralNetwork {
    fn new(vs: VarBuilder) -> Result<Self, Box<dyn Error>> {
        let ln1 = linear(10, 100, vs.pp("ln1"))?;
        let ln2 = linear(100, 1, vs.pp("ln2"))?;
        Ok(Self { ln1, ln2 })
    }
}

impl Module for DenseNeuralNetwork {
    fn forward(&self, xs: &Tensor) -> candle::Result<Tensor> {
        let xs = self.ln1.forward(xs)?;
        let xs = xs.relu()?;
        self.ln2.forward(&xs)
    }
}

fn main() -> Result<(), Box<dyn Error>> {
    let device = Device::cuda_if_available(0)?;
    println!("Using device: {:?}", device);
    let varmap = VarMap::new();
    let vs = VarBuilder::from_varmap(&varmap, DType::F32, &device);

    let dnn = DenseNeuralNetwork::new(vs)?;
    let mut optimizer =
        AdamW::new(varmap.all_vars(), ParamsAdamW::default())?;

    let beta = Tensor::randn(0f32, 1., (10, 1), &device)?;

    // Generate a training set
    let xs_train = Tensor::randn(0f32, 1., (100, 10), &device)?;
    let mu = xs_train.cos()?.matmul(&beta)?;
    let eps = Tensor::randn(0f32, 1., (100, 1), &device)?;
    let ys_train = (mu + eps)?;

    // Generate a validation set
    let xs_val = Tensor::randn(0f32, 1., (100, 10), &device)?;
    let mu = xs_val.cos()?.matmul(&beta)?;
    let eps = Tensor::randn(0f32, 1., (100, 1), &device)?;
    let ys_val = (mu + eps)?;

    let n_epochs = 10_000;
    let mut losses_val = Vec::<f32>::with_capacity(n_epochs);

    for epoch in tqdm(0..n_epochs) {
        if epoch % (n_epochs / 10) == 0 || epoch == n_epochs - 1 {
            losses_val
                .push(mse(&dnn.forward(&xs_val)?, &ys_val)?.to_scalar()?);
        }

        let gradients =
            mse(&dnn.forward(&xs_train)?, &ys_train)?.backward()?;
        optimizer.step(&gradients)?;
    }
    println!("Losses on validation set: {:?}", losses_val);
    Ok(())
}
